from SkitterCrawler.util import AuthHandler, FileChecker, TimeWaiter, Log, run
from SkitterCrawler.twitter_crawlers import RandomWalkerTwitterCrawler
from pathlib import Path
from functools import partial
import json
import os


def _random_walk_crawler(config_m=None):
    if config_m is None:
        skittercrawler = os.path.join(str(Path.home()), "SkitterCrawler")
        f = open(os.path.join(skittercrawler, "rwc_config.json"), 'r')
        config = json.load(f)
        f.close()
    elif type(config_m) is not dict:
        f = open(config_m, 'r')
        config = json.load(f)
        f.close()
    else:
        config = config_m

    crawlers, crawlers_names = [], list(config.keys())
    for crawler_name in crawlers_names:
        auth_handler = AuthHandler()
        auth_dict = auth_handler.get_crawler_config(crawler_name)
        crawler = RandomWalkerTwitterCrawler(auth_dict=auth_dict,
                                             stopper=partial(TimeWaiter, config[crawler_name]['time_wait']),
                                             logger=Log,
                                             seed=config[crawler_name]['seed'])
        crawlers.append(crawler.run)

    state = RandomWalkerTwitterCrawler.init_external_state
    reducer = RandomWalkerTwitterCrawler.reducer

    run(slaves=crawlers, identifiers=crawlers_names, reduce=reducer, check=FileChecker,  init_state=state)


if __name__ == "__main__":
    _random_walk_crawler()



from pathlib import Path
import sqlalchemy.orm
import sqlalchemy
import threading
import datetime
import pickle
import time
import json
import os


class Memory:

    def __init__(self, attrs=[], attrs_keep=[],  local=False):
        self.track = dict()
        for i in attrs:
            self.add_attribute(i)
        self.attrs_keep = attrs_keep
        self.dir = os.path.join(str(Path.home()), "SkitterCrawler")
        self.pickled = os.path.join(self.dir, "mem.p")

        if os.path.exists(self.pickled) and not local:
            f = open(self.pickled, "rb")
            other = pickle.load(f)
            f.close()
            self.track = other.track
            self.attrs_keep = other.attrs_keep

    def add_attribute(self, attribute):
        self.track[attribute] = dict()

    def add_val(self, attribute, val, ret):
        self.track[attribute][val] = ret

    def is_attribute(self, attribute):
        return attribute in self.track

    def is_val(self, attribute, val):
        return val in self.track[attribute]

    def get_attribute(self, attribute):
        return self.track[attribute]

    def get_val(self, attribute, val):
        return self.track[attribute][val]

    def is_in_memory(self, attribute, val):
        return self.is_attribute(attribute) and self.is_val(attribute, val)

    def dump(self):
        f = open(self.pickled, "wb")
        pickle.dump(self, f)
        f.close()


class DBHandler:
    """ This handles the operations with the database. It:
        1) Creates SQLAlchemy engine for some database.
        2) Start/closes new sessions for this database.
        3) Creates or drop ORM schemas for the database.
        4) Get values from the file with get_config/get_db_config.
    """

    def __init__(self, base):

        self.auth_handler = AuthHandler()
        self.engine = sqlalchemy.create_engine(self.auth_handler.get_db_config())
        self.sessionmaker = sqlalchemy.orm.sessionmaker(bind=self.engine)
        self.session = None
        self.base = base

    def start_new_session(self):

        self.sessionmaker.close_all()
        self.session = self.sessionmaker()

    def close_session(self):

        self.session.close()
        self.session = None

    def is_section_active(self):

        return self.session is not None

    def create_base(self):

        self.base.metadata.create_all(self.engine)

    def drop_base(self):

        self.base.metadata.drop_base(self.engine)

    def change_db_config(self, config):

        self.auth_handler.change_db_config(config)

    def get_db_config(self):

        return self.auth_handler.get_db_config()

    def commit_session(self):
        self.session.commit()


class AuthHandler:
    """ This handles the authentication with the api/databases, it:
        1) Creates a new file for secrets in __init__, if there isn't.
        2) Attach values to the file with attach_update_config/change_db_config.
        3) Remove values from the file with pop_config.
        4) Get values from the file with get_config/get_db_config.
    """

    def __init__(self):

        self.dir = os.path.join(str(Path.home()), "SkitterCrawler")
        self.json_file = os.path.join(self.dir, "secrets.json")
        self.db_address = os.path.join(self.dir, "db.txt")

        if not os.path.exists(self.dir):
            os.makedirs(self.dir)

        if not os.path.exists(self.json_file):
            f = open(self.json_file, 'w')
            f.write(json.dumps(dict()))
            f.close()

        if not os.path.exists(self.db_address):
            f = open(self.db_address, 'w')
            f.write('')
            f.close()

    def attach_crawler_config(self, key, value, crawler=None):

        f = open(self.json_file, 'r')
        tmp = json.loads(f.read())
        if crawler is None:
            tmp[key] = value
        else:
            if crawler not in tmp:
                tmp[crawler] = dict()
            tmp[crawler][key] = value
        f.close()
        f = open(self.json_file, 'w')
        f.write(json.dumps(tmp))
        f.close()

    def pop_crawler_config(self, key, crawler=None):

        f = open(self.json_file, 'r')
        tmp = json.loads(f.read())
        v = None
        if crawler is None:
            v = tmp.pop(key, None)
        else:
            if crawler not in tmp:
                v = tmp[crawler].pop(key, None)
        f.close()
        f = open(self.json_file, 'w')
        f.write(json.dumps(tmp))
        f.close()
        return v

    def get_crawler_config(self, key=None, crawler=None):

        f = open(self.json_file, 'r')
        tmp = json.loads(f.read())
        f.close()
        if key is None:
            return tmp
        elif crawler is None:
            return tmp[key]
        else:
            return tmp[crawler][key]

    def change_db_config(self, config):

        f = open(self.db_address, 'w')
        f.write(config)
        f.close()

    def get_db_config(self):

        f = open(self.db_address, 'r')
        tmp = f.read()
        f.close()
        return tmp


class FileChecker:
    """ This is the default class for the Checker, it:
        1) Creates a file with a flag 1 on `__init__'.
        2) On `start' sets the flag to 1.
        3) On `check' returns whether the flag is 1.
        4) On `stop' sets the flag to 0.
    """

    def __init__(self):
        self.dir = os.path.join(str(Path.home()), "SkitterCrawler")
        self.txt = os.path.join(self.dir, "._control.txt")
        if not os.path.exists(self.dir):
            os.makedirs(self.dir)
        self.start()

    def start(self):
        f = open(self.txt, 'w')
        f.write("1")
        f.close()

    def check(self):
        f = open(self.txt, 'r')
        tmp = f.read() == "1"
        f.close()
        return tmp

    def stop(self):
        f = open(self.txt, 'w')
        f.write("0")
        f.close()


class TimeWaiter:
    """ This is the default class for the Stopper, it:
        1) Sets a number of seconds in `__init__' or on `start'.
        2) Starts the timer on `start'.
        3) Checks if the timer is over on on `check'
    """

    def __init__(self, default=None):

        self.hammer_time = 0
        self.default = 60 if default is None else default

    def start(self, seconds=None):

        t = self.default if seconds is None else seconds
        self.hammer_time = (datetime.datetime.now() + datetime.timedelta(seconds=int(t))).timestamp()

    def check(self):

        return datetime.datetime.now().timestamp() < self.hammer_time


class Log:
    """ This is the default class for the Logger, it:
        1) Starts a dict() on `__init__'.
        2) `log' creates if key not in dict, creates dict[string] = [], else dict[string].append(list_stuff).
        3) `ret' returns the lists sorted by time
    """

    def __init__(self):
        self.dir = os.path.join(str(Path.home()), "SkitterCrawler")
        self.txt_file = os.path.join(self.dir, "log.txt")
        self.info = dict()
        self.id = "(noid)"

    def set_id(self, id_v):
        self.id = id_v

    def log(self, string, list_stuff):

        if string not in self.info:
            self.info[string] = []
        for stuff in list_stuff:
            self.info[string].append("{0} {1}: {2}".format(self.id,
                                                        time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime()),
                                                        stuff))

    def __add__(self, other):
        for key in self.info:
            if key not in other.info:
                other.info[key] = []
        for key in other.info:
            if key not in self.info:
                self.info[key] = []
        for key in self.info:
            self.info[key] = self.info[key] + other.info[key]
        self.id = "(noid)"

        return self

    def ret(self):

        string = ""

        for key in sorted(self.info.keys()):
            string += key + "\n"
            for stuff in self.info[key]:
                string += "\t" + stuff + "\n"

        return string

    def dump(self, s=None):
        f = open(self.txt_file, 'a+')
        if s is not None:
            f.write(s)
        f.write(self.ret())
        f.close()

    def flush(self):
        self.info = dict()
        self.id = "(noid)"


def run(slaves, identifiers, reduce, check=FileChecker, init_state=lambda: None):
    """ This function coordinates data collection, distributing tasks to several slaves and then merging the results.
    1) Each slave is a runnable function that does something we want. It must contain two arguments to pass on results.
    2) Each slave is assigned a unique identifier.
    3) After all slaves finish their execution, the reduce function process the work of all slaves.
    4) The cycle then continues forever.

    :param slaves: list of runnable functions that take a string, a list and an index as arguments.
    :param identifiers: list of identifiers.
    :param reduce: function that receives list of identifiers and list of results.
    :param check: function that yields true if the cycle is to continue.
    :param init_state: function to initialize shared state.
    :return: number of iterations.
    """

    it = 0
    state = init_state()

    checker = check()
    checker.start()

    while checker.check():

        l = len(slaves)
        results = [None] * l
        threads = [None] * l

        for i, s, idx in zip(identifiers, slaves, range(l)):
            threads[idx] = threading.Thread(name=i, target=s, args=(i, results, idx, state))

        for idx in range(l):
            threads[idx].setDaemon(True)

        for idx in range(l):
            threads[idx].start()

        for idx in range(l):
            threads[idx].join()

        state = reduce(it=it, results=results, state=state)
        it += 1

    return it



from setuptools import setup

setup(name='SkitterCrawler',
      version='0.0.2',
      description='Generic packages with Crawlers for OSNs',
      url='https://github.com/manoelhortaribeiro/SkitterCrawler',
      download_url='https://github.com/manoelhortaribeiro/SkitterCrawler',
      author='Manoel Horta Ribeiro',
      author_email='manoelhortaribeiro@gmail.com',
      license='MIT',
      packages=['SkitterCrawler'],
      keywords='Online social network crawler utilities',
      install_requires=['sqlalchemy', 'SkitterSchema', 'tweepy', 'numpy'],
      entry_points={
            'console_scripts': ['rw_crawler=SkitterCrawler.random_walker_crawler:_random_walk_crawler']
      },
      include_package_data=True,
      zip_safe=False)



from SkitterSchema.tweepy2schema import ProcessedOutcome, process_timeline, process_user, process_followers
from SkitterSchema.tweepy2schema import process_followees, process_favourites
from SkitterCrawler.util import Memory, Log, DBHandler
from SkitterSchema.twitter_schema import UUEdge
from SkitterSchema.twitter_schema import Base
import random
import tweepy


class TwitterCrawler:
    def __init__(self, auth_dict, stopper, logger):
        """ Initialization method for the generic crawler.
        :param auth_dict: dictionary with consumer_key, consumer_secret, access_token and access_secret for twitter api.
        :param logger: logger class with log.log(key, values)
        :param stopper: stopper class with stopper.start() and stopper.check()
        """

        # gets secrets
        self.consumer_key, self.consumer_secret = auth_dict['consumer_key'], auth_dict['consumer_secret']
        self.access_token, self.access_secret = auth_dict['access_token'], auth_dict['access_secret']

        # initializes logger and stopper
        self.log, self.stop = logger(), stopper()

        # gets api object from Tweepy
        auth = tweepy.OAuthHandler(self.consumer_key, self.consumer_secret)
        auth.set_access_token(self.access_token, self.access_secret)
        self.api = tweepy.API(auth)

        self.log.log("Initializing API", ["API initialization was successful"])

    def _run(self, identifier, result, idx, state, **kwargs):
        raise Exception("_run function not implemented!")

    def run(self, identifier, result, idx, state, **kwargs):
        """ This function is a generic slave.
        :param identifier: identifier for the slave.
        :param result: array with results.
        :param idx: position of the array that can be modified.
        :param state: state of the current iteration.
        :param kwargs: arguments for filtering/searching.
        :return: Nothing
        """
        self._run(identifier, result, idx, state, **kwargs)

    @staticmethod
    def reducer(it, results, state):
        raise Exception("reducer function not implemented!")


class RandomWalkerTwitterCrawler(TwitterCrawler):
    def __init__(self, auth_dict, stopper, logger, seed, prob_types=(0.7, 0.3), prob_seed=0.3):
        """ This initializes a RandomWalkerTwitterCrawler.
        :param auth_dict: dictionary with consumer_key, consumer_secret, access_token and access_secret for twitter api.
        :param logger: logger class with log.log(key, values).
        :param stopper: stopper class with stopper.start() and stopper.check().
        :param seed: seed or list of seeds from which the rw will start and eventually return.
        :param prob_types: P of next u_retweeted_by_me, u_replied_by_me.
        :param prob_seed: probability at each hop that rw will return to the seed.
        """
        self.seed, self.pointer = seed, None  # Initializes seed(s), pointer
        self.prob, self.prob_seed = [sum(prob_types[:i + 1]) for  # Initializes probabilities
                                     i, x in enumerate(prob_types)], prob_seed
        self.excd, self.count, self.got, self.mem = dict(), dict(), dict(), None  # Bookkeeping
        super().__init__(auth_dict, stopper, logger)

    def point_to_seed(self):
        """ This makes the pointer points to the seed. If the seed is unique, it will always point to the same seed,
        however if the seed is a list of tweet IDs, we return a seed randomly.
        :return: Nothing
        """
        self.pointer = random.choice(self.seed) if isinstance(self.seed, list) else self.seed

    def chooses_next(self, retweets, replies):
        """ Chooses next from retweets, replies or the existing seed.
        :param retweets: list of user ids of whom the user retweeted.
        :param replies: list of user ids of whom the user replied.
        :return: Nothing.
        """
        joint = [retweets, replies]

        rv = random.random()
        if rv <= self.prob_seed:
            self.point_to_seed()
            return

        rv = random.random()
        tmp = [rv <= i for i in self.prob].index(True)
        order = list(range(tmp, len(joint))) + list(range(0, tmp))

        for i in order:
            if len(joint[i]) == 0:
                continue
            self.pointer = random.choice(joint[i])
            return

        self.point_to_seed()

    def get_user(self):
        """ This makes an api call to get the user and then processes it. """
        return process_user(self.api.get_user(self.pointer))

    def get_timeline(self):
        """ This makes an api call to get the timeline and then processes it. """
        retweets, replies, timeline_outcome = [], [], ProcessedOutcome()
        tl = self.api.user_timeline(include_rts=True, count=200, trim_user=False, exclude_replies=False,
                                    user_id=self.pointer)
        timeline_outcome += process_timeline(tl)
        retweets = set(timeline_outcome.get_filtered_edges(UUEdge, "retweet", lambda x: x.dst_id))
        retweets = list(filter(lambda x: x != self.pointer, retweets))
        replies = set(timeline_outcome.get_filtered_edges(UUEdge, "reply", lambda x: x.dst_id))
        replies = list(list(filter(lambda x: x != self.pointer, replies)))
        return retweets, replies, timeline_outcome

    def get_followers_ees(self):
        """ This makes an api call to get the followers/ees and then processes it. """
        followees = self.api.friends_ids(user_id=self.pointer, count=5000)
        followers = self.api.followers_ids(user_id=self.pointer, count=5000)
        followers_ees_outcome = process_followers(followers, self.pointer)
        followers_ees_outcome += process_followees(followees, self.pointer)
        return followers_ees_outcome

    def get_favs(self):
        """ This makes an api call to get the favorites and then processes it. """
        favs = self.api.favorites(user_id=self.pointer, count=200, include_entities=True)
        favs_outcome = process_favourites(favs, self.pointer)
        return favs_outcome

    def double_mem_check(self, state, attribute, val, func):
        """This executes the given func only if the val is not in self.memory[attribute]"""
        outcome = ProcessedOutcome()
        # Gets user if not in memory, may raise tweepy.RateLimitError
        if not self.mem.is_in_memory(attribute, val) and not state.is_in_memory(attribute, val):
            outcome = func()
            self.count[attribute] += 1

            if self.mem.is_attribute(attribute):
                self.mem.add_val(attribute, val, True)
        self.got[attribute] = True
        return outcome

    def handle_got(self, e, attribute):
        if not self.got[attribute]:
            self.log.log("Error Retrieving {0}".format(attribute), [e.reason])
            if type(e) is tweepy.RateLimitError:
                self.excd[attribute] = True
                return True
        return False

    def _run(self, identifier, result, idx, state, **kwargs):
        """ Internal function for doing the streaming processing. """

        self.stop.start()
        self.point_to_seed()
        self.log.flush()
        self.log.set_id(identifier)
        outcome = ProcessedOutcome()
        retweets, replies = [], []
        self.mem = Memory(['user', 'tl', 'flwer', 'favs'], ['flwer', 'favs'], local=True)
        self.count = dict({"user": 0, "tl": 0, "flwer": 0, "favs": 0})
        self.excd = dict({"user": False, "tl": False, "flwer": False, "favs": False})

        while self.stop.check() and self.excd['tl'] is False:

            self.got = dict({"user": False, "tl": False, "flwer": False, "favs": False})

            try:
                # Gets timeline if not in memory, may raise tweepy.RateLimitError
                if self.mem.is_in_memory('tl', self.pointer):
                    retweets, replies = self.mem.get_val('tl', self.pointer)
                elif state.is_in_memory('tl', self.pointer):
                    retweets, replies = state.get_val('tl', self.pointer)
                else:
                    retweets, replies, timeline_outcome = self.get_timeline()
                    outcome += timeline_outcome
                    self.count['tl'] += 1
                    if self.mem.is_attribute('tl'):
                        self.mem.add_val('tl', self.pointer, [retweets, replies])
                self.got['tl'] = True

                # Gets user if not in memory, may raise tweepy.RateLimitError
                if not self.excd["user"]:
                    outcome += self.double_mem_check(state, 'user', self.pointer, self.get_user)

                # Gets followers if not in memory, may raise tweepy.RateLimitError
                if not self.excd["flwer"]:
                    outcome += self.double_mem_check(state, 'flwer', self.pointer, self.get_followers_ees)

                # Gets followers if not in memory, may raise tweepy.RateLimitError
                if not self.excd["favs"]:
                    outcome += self.double_mem_check(state, 'favs', self.pointer, self.get_favs)

            except tweepy.TweepError as e:

                if self.handle_got(e, 'tl'):
                    continue

                if self.handle_got(e, 'user'):
                    continue

                if self.handle_got(e, 'flwer'):
                    continue

                if self.handle_got(e, 'favs'):
                    continue

            previous_pointer = self.pointer
            self.chooses_next(retweets, replies)
            print("({0}) {1} > {2} : user={3}/{7}/{11} tl={4}/{8}/{12} flwer={5}/{9}/{13} favs={6}/{10}/{14}".format(
                identifier, previous_pointer, self.pointer,
                self.count["user"], self.count["tl"], self.count["flwer"], self.count["favs"],
                len(self.mem.track["user"]), len(self.mem.track["tl"]),
                len(self.mem.track["flwer"]), len(self.mem.track["favs"]),
                len(state.track["user"]), len(state.track["tl"]),
                len(state.track["flwer"]), len(state.track["favs"])))

            retweets, replies = [], []

        self.log.log("Retrieved", ["users: {0}, statuses: {1}, entities: {2}, edges: {3}"
                     .format(len(outcome.schema_users), len(outcome.schema_tweets),
                             len(outcome.schema_entities), len(outcome.schema_edges))])

        self.log.log("Requests", ["user: {0}, tl: {1}, flwer: {2}, favs: {3}"
                     .format(self.count["user"], self.count["tl"], self.count["flwer"], self.count["flwer"])])

        self.log.log("Memory", ["Local user: {0}, tl: {1}, flwer: {2}, favs: {3}"
                     .format(len(self.mem.track["user"]), len(self.mem.track["tl"]),
                             len(self.mem.track["flwer"]), len(self.mem.track["favs"]))])

        self.log.log("Memory", ["Global user: {0}, tl: {1}, flwer: {2}, favs: {3}"
                     .format(len(state.track["user"]), len(state.track["tl"]),
                             len(state.track["flwer"]), len(state.track["favs"]))])

        result[idx] = (identifier, outcome, self.mem, self.log)

    @staticmethod
    def reducer(it, results, state):

        nlog = Log()
        dbhandler = DBHandler(base=Base)
        dbhandler.create_base()
        dbhandler.start_new_session()
        final_outcome = ProcessedOutcome()

        state.dump()

        for identifier, outcome, state_update, log in results:
            final_outcome += outcome
            for attr in state_update.attrs_keep:
                for val in state_update.get_attribute(attr).keys():
                    state.add_val(attr, val, True)
            nlog += log

        final_outcome.add_to_session(dbhandler.session)
        dbhandler.commit_session()
        dbhandler.close_session()
        nlog.dump(s="Iteration {0}:\n".format(it))

        return state

    @staticmethod
    def init_external_state():
        return Memory(['user', 'tl', 'flwer', 'favs'], ['flwer', 'favs'])
